use crate::config::MCTSConfig;
use crate::policy::backprop::backpropagate_rewards;
use crate::policy::decision::decide_on_action;
use crate::policy::selection::select_best_child;
use crate::policy::simulation::simulate_reward;
use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::{Node, RewardVal};
use crate::tree::print_tree;
use rand::prelude::IndexedRandom;
use std::collections::HashMap;
use std::time::Instant;

/// Monte Carlo Tree Search implementation
///
/// This provides the interface for performing optimal searches on a tree using
/// the MCTS algorithm.
pub struct MCTS<'conf, S: GameState> {
    /// The arena used for the tree
    arena: Arena<S>,

    /// The identifier of the root node of the search tree
    root_id: usize,

    /// The configuration used for the search
    config: &'conf MCTSConfig<S>,
}

impl<'conf, S: GameState + std::fmt::Debug> MCTS<'conf, S> {
    /// Creates a new instance with the given initial state and configuration
    pub fn new(initial_state: S, config: &'conf MCTSConfig<S>) -> Self {
        let mut arena: Arena<S> = Arena::new(config.tree_size_allocation);
        let root: Node<S> = Node::new(initial_state.clone(), None, None, 0);
        let root_id: usize = arena.add_node(root);
        MCTS {
            arena,
            root_id,
            config,
        }
    }

    /// Runs the MCTS algorithm, returning the "best" action
    ///
    /// The search will stop once `max_iterations` or `max_time` from
    /// the assigned configration is reached.
    pub fn search(&mut self) -> Result<S::Action> {
        self.search_for_iterations(self.config.max_iterations)
    }

    /// Runs the MCTS algorithm, returning the "best" action after the given iterations
    ///
    /// This ignores the `max_iterations` provided in the config, however will
    /// return if `max_time` is specific and reached before the iterations are complete.
    pub fn search_for_iterations(&mut self, iterations: usize) -> Result<S::Action> {
        let start_time = Instant::now();
        for _ in 0..iterations {
            match self.config.max_time {
                Some(max_time) => {
                    if start_time.elapsed() >= max_time {
                        break; // ending early due to time
                    }
                }
                None => {}
            }
            self.execute_iteration()?;
        }

        self.best_action()
    }

    /// Runs the MCTS algorithm for a single iteration
    fn execute_iteration(&mut self) -> Result<()> {
        let mut selected_id: usize = self.select();
        let selected_node: &Node<S> = self.arena.get_node(selected_id);
        if !selected_node.state.is_terminal() {
            self.expand(selected_id);
            let children: &Vec<usize> = &self.arena.get_node(selected_id).children;
            match children.choose(&mut rand::rng()) {
                Some(&random_child) => {
                    selected_id = random_child;
                }
                None => {
                    // We ran out of nodes
                    return Err(MCTSError::NonTerminalGame);
                }
            }
        }
        let rewards = self.simulate(selected_id);
        self.backprop(selected_id, &rewards);

        Ok(())
    }

    /// MCTS Phase 1: Selection - Find the "best" node to expand
    fn select(&mut self) -> usize {
        let mut current_id: usize = self.root_id;
        loop {
            let node = &self.arena.get_node(current_id);
            if node.is_leaf() || node.state.is_terminal() {
                return current_id;
            }
            current_id = select_best_child(&self.config.selection_policy, &node, &self.arena);
        }
    }

    /// MCTS Phase 2: Expansion - Expand the selected node on the tree
    fn expand(&mut self, id: usize) {
        let parent: &Node<S> = self.arena.get_node(id);
        let legal_actions: Vec<S::Action> = parent.state.get_legal_actions();
        let parent_state: S = parent.state.clone();
        let depth: usize = parent.depth + 1;
        for action in legal_actions {
            let state = parent_state.state_after_action(&action);
            let new_node = Node::new(state, Some(action), Some(id), depth);
            let new_id = self.arena.add_node(new_node);
            self.arena.get_node_mut(id).children.push(new_id);
        }
    }

    fn simulate(&self, id: usize) -> HashMap<S::Player, RewardVal> {
        let node = &self.arena.get_node(id);
        simulate_reward(&self.config.simulation_policy, &node, &self.arena)
    }

    fn backprop(&mut self, selected_id: usize, rewards: &HashMap<S::Player, RewardVal>) {
        backpropagate_rewards(
            &self.config.backprop_policy,
            selected_id,
            &mut self.arena,
            &rewards,
        )
    }

    fn best_action(&self) -> Result<S::Action> {
        let root_node: &Node<S> = self.arena.get_node(self.root_id);
        match decide_on_action(&self.config.decision_policy, &root_node, &self.arena) {
            Some(action) => Ok(action),
            None => Err(MCTSError::NoBestAction),
        }
    }

    pub fn print_tree(&self) {
        print_tree(self.root_id, &self.arena)
    }
}

/// Errors returned by the MCTS algorithm
#[derive(Debug, thiserror::Error)]
pub enum MCTSError {
    /// The best action doesn't exist
    #[error("Unable to determine a best action for the game")]
    NoBestAction,

    /// The search tree was exhausted without finding a terminal node
    #[error("Search tree exhausted without finding terminal node")]
    NonTerminalGame,
}

/// Result returned by the MCTS algorithm
pub type Result<T> = std::result::Result<T, MCTSError>;
