use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::RewardVal;
use std::collections::HashMap;

/// The back propagation policy dictating the propagation of playout results
///
/// This policy drives how the backpropagation phase of the MCTS algorithm is
/// executed, allowing for some minor customization.
///
/// Typically the Standard policy, used by most implementaions of MCTS, is
/// sufficient
#[derive(Debug)]
pub enum BackpropagationPolicy<S: GameState> {
    /// Standard back propagation
    ///
    /// This increments the visitation count and adds the simulated rewards
    /// results to the aggregate values.
    ///
    /// This is the standard policy used in most MCTS implementations.
    Standard,

    /// Weighted back propagation
    ///
    /// This weights the value of the simulated rewards based on the depth,
    /// allowing us to put more-or-less influence on deeper branches
    /// - Positive weight factor makes deeper nodes less influential
    /// - Negative weight factor makes deeper nodes more influential
    Weighted(f64),

    /// Custom backpropagation policy
    Custom(Box<dyn CustomBackpropagationPolicy<S>>),
}

/// Trait for an object implementing the backpropagation logic whene exploring the MCTS
/// search tree.
pub trait CustomBackpropagationPolicy<S: GameState>: std::fmt::Debug {
    /// Backpropagate the given rewards values from the node up the tree
    fn backprop(
        &self,
        node_id: usize,
        arena: &mut Arena<S>,
        rewards: &HashMap<S::Player, RewardVal>,
    );
}

pub fn backpropagate_rewards<S: GameState>(
    policy: &BackpropagationPolicy<S>,
    node_id: usize,
    arena: &mut Arena<S>,
    rewards: &HashMap<S::Player, RewardVal>,
) {
    match policy {
        BackpropagationPolicy::Standard => standard_backprop(node_id, arena, rewards),
        BackpropagationPolicy::Weighted(depth_factor) => {
            weighted_backprop(*depth_factor, node_id, arena, rewards)
        }
        BackpropagationPolicy::Custom(custom_policy) => {
            custom_policy.backprop(node_id, arena, rewards)
        }
    }
}

fn standard_backprop<S: GameState>(
    node_id: usize,
    arena: &mut Arena<S>,
    rewards: &HashMap<S::Player, RewardVal>,
) {
    let mut current_id: usize = node_id;
    loop {
        let node = arena.get_node_mut(current_id);
        node.increment_visits();
        for (player, reward) in rewards.iter() {
            node.record_player_reward(player.clone(), *reward);
        }
        if let Some(parent_id) = node.parent {
            current_id = parent_id;
        } else {
            break;
        }
    }
}

fn weighted_backprop<S: GameState>(
    depth_factor: f64,
    node_id: usize,
    arena: &mut Arena<S>,
    rewards: &HashMap<S::Player, RewardVal>,
) {
    let mut current_id: usize = node_id;
    loop {
        let node = arena.get_node_mut(current_id);
        let weight = weight_for_depth(depth_factor, node.depth);
        for (player, reward) in rewards.iter() {
            node.record_player_reward(player.clone(), (*reward) * weight);
        }
        if let Some(parent_id) = node.parent {
            current_id = parent_id;
        } else {
            break;
        }
    }
}

/// Calculate the weight based on the current depth
///
/// The weight can be multiplied agains the reward to provide more or less
/// influence based on depth as determined by the depth factor
fn weight_for_depth(depth_factor: f64, depth: usize) -> f64 {
    1.0 / (1.0 + depth_factor * depth as f64)
}
