mod ucb1;
mod ucb1_tuned;

use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::Node;

/// The selection policy used in the MCTS selection phase
///
/// This drives the selection of the nodes in the search tree, determining
/// which paths are explored and evaluated.
///
/// In general UCB1-Tuned or UCB1 should be effective, however if necessariy
/// a custom selection policy can be provided.
#[derive(Debug)]
pub enum SelectionPolicy<S: GameState> {
    /// Upper Confidence Bound 1 (UCB1) with the given exploration constant
    ///
    /// The exploration constant controls the balance between exploration and
    /// exploitation. The higher the value, the mroe likely the search will
    /// explore less-visited nodes. A standard value is √2 ≈ 1.414.
    UCB1(f64),

    /// Upper Confidence Bound 1 Tuned (UCB1-Tuned)
    ///
    /// A tuned version of UCB1 instead using the empirical
    /// standard deviation of the rewards to drive exploration.
    ///
    /// Auer, P., Cesa-Bianchi, N. & Fischer, P. Finite-time Analysis of the Multiarmed Bandit Problem. Machine Learning 47, 235–256 (2002). https://doi.org/10.1023/A:1013689704352
    UCB1Tuned(f64),

    /// Custom selection policy
    Custom(Box<dyn CustomSelectionPolicy<S>>),
}

/// Trait for an object implementing the selection logic whene exploring the MCTS
/// search tree.
///
/// The policy should select the child of the given node which is "best" for the current player
pub trait CustomSelectionPolicy<S: GameState>: std::fmt::Debug {
    /// Selects a child based on the policy, returning the node ID
    fn select_child(&self, node: &Node<S>, arena: &Arena<S>) -> usize;
}

pub fn select_best_child<S: GameState>(
    policy: &SelectionPolicy<S>,
    node: &Node<S>,
    arena: &Arena<S>,
) -> usize {
    match policy {
        SelectionPolicy::UCB1(exploration_constant) => {
            ucb1::select_best_child(*exploration_constant, node, arena)
        }
        SelectionPolicy::UCB1Tuned(exploration_constant) => {
            ucb1_tuned::select_best_child(*exploration_constant, node, arena)
        }
        SelectionPolicy::Custom(custom_policy) => custom_policy.select_child(node, arena),
    }
}
