//! Upper Confidence Bound 1 (UCB1) selection policy
//!
//! This is the classic selection policy for MCTS, which balances
//! exploration and exploitation using the UCB1 formula:
//!
//! ```text
//! UCB1 = average_reward + exploration_constant * sqrt(ln(parent_visits) / child_visits)
//! ```
//!
//! Where:
//! - `average_reward` is the average reward from simulations through this node
//! - `exploration_constant` controls the balance between exploration and exploitation
//! - `parent_visits` is the number of visits to the parent node
//! - `child_visits` is the number of visits to the child node
//!
//! Higher exploration constants favor exploration (trying less-visited nodes),
//! while lower values favor exploitation (choosing nodes with higher values).
//!
//! The commonly used value for the exploration constant is sqrt(2) ≈ 1.414,
//! which is the default in this implementation.

use crate::state::GameState;
use crate::tree::arena::Arena;
use crate::tree::node::{Node, RewardVal};

/// Selects the index of the "best" child using the UCB1 selection policy
pub fn select_best_child<S: GameState>(
    exploration_constant: f64,
    node: &Node<S>,
    arena: &Arena<S>,
) -> usize {
    if node.is_leaf() {
        panic!("select_best_child called on leaf node");
    }

    let player = node.state.get_current_player();
    let parent_visits = node.visits;
    let best_child = node
        .children
        .iter()
        .max_by(|&a, &b| {
            let node_a = arena.get_node(*a);
            let node_b = arena.get_node(*b);
            let ucb_a = ucb1_value(
                exploration_constant,
                node_a.reward_average(player),
                node_a.visits,
                parent_visits,
            );
            let ucb_b = ucb1_value(
                exploration_constant,
                node_b.reward_average(player),
                node_b.visits,
                parent_visits,
            );
            ucb_a.partial_cmp(&ucb_b).unwrap()
        })
        .unwrap();
    *best_child
}

/// Calculates the UCB1 value for a node
pub fn ucb1_value(
    exploration_constant: f64,
    child_value: RewardVal,
    child_visits: u64,
    parent_visits: u64,
) -> RewardVal {
    if child_visits == 0 {
        return f64::INFINITY; // Always explore nodes that have never been visited
    }

    // UCB1 formula: value + C * sqrt(ln(parent_visits) / child_visits)
    let exploitation = child_value;
    let exploration =
        exploration_constant * ((parent_visits as f64).ln() / child_visits as f64).sqrt();

    exploitation + exploration
}
